==> untahris-2.0-src.tgz: untahris/net.cpp <==
// Untahris (Common Playground)
// copyright (C) 2006 longuens@users.sourceforge.net
// Released under GNU General Public License (see the file COPYING)
// --------------------------------------------------------------
// Network utilities.

// The algorithm for netplay is as follows. Each keypress (or join) is
// stored in the key queue, together with its time. These are sent
// across the network. Tha playground acts deterministically given
// these keypresses. Backups of playground are saved, so if a keypress
// arrives when the playground is already newer than its time, we can
// retrace to the old playground.

#include <SDL/SDL_net.h>

// struct to be sent via net, and for storing backups
struct TBuffer {
  int readpos;
  vector<char> buf;
  
  template<class T> void put(T t) {
    const int s = sizeof(T);
    char *p = (char*) &t;
    for(int i=0; i<s; i++) buf.push_back(p[i]);
    }

  void putChar(char c) { put(c); }
  void putInt(int32 c) { put(c); }
  
  int getInt() { int32 i; get(i); return i;}
  
  void putC(CellType c) { putChar((char) c); }
  void getC(CellType& c) {char ch; get(ch); c = CellType(ch); }
  
  template<class T> void get(T& t) {
    if(readpos >= size(buf)) {
      printf("Protocol error\n");
      exit(3);
      }
    const int s = sizeof(T);
    char *p = (char*) &t;
    for(int i=0; i<s; i++) p[i] = buf[readpos++];
    }
  
  TPlayObject *getObject();
  
  bool send(TCPsocket skt);
  bool receive(TCPsocket skt);
  
  int getTime() {
    return *((int*)(&(buf[0])));
    }

  int getKQPos() {
    return *((int*)(&(buf[4])));
    }
  
  void appendTo(TBuffer& b) {
    for(int i=0; i<size(buf); i++) b.putChar(buf[i]);
    }
  };

#define QSAVES 32
#define SAVEPERIOD 500

int lastSave;
int firstSave;

TBuffer saves[QSAVES];

bool TBuffer::send(TCPsocket skt) {
  int result;
  
  int len = size(buf);
  
  result = SDLNet_TCP_Send(skt, &len, 4);
  if(result != 4) return false;

  result = SDLNet_TCP_Send(skt, &(buf[0]), len);
  if(result != len) return false;
  return true;
  }

bool TBuffer::receive(TCPsocket skt) {
  int result, len;

  result = SDLNet_TCP_Recv(skt, &len, 4);
  if(result != 4) return false;
  buf.resize(len);

  int pos = 0;
  while(pos < len) {
    result = SDLNet_TCP_Recv(skt, &(buf[pos]), len-pos);
    if(result < 1) return false;
    pos += result;
    }

  readpos = 0;
  return true;
  }

void TPlayground::save(TBuffer& b) {
  b.put(time);
  b.put(kqpos);
  b.put(seed);
  b.put(sizeX);
  b.put(sizeY);
  
  for(int y=0; y<sizeY; y++) for(int x=0; x<sizeX; x++) {
    b.putC(map[y][x]);
    b.put(owner[y][x]);
    }
  
  b.putInt(size(objects));
  for(int o=0; o<size(objects); o++) objects[o]->save(b);

  b.putInt(size(pl));
  for(int i=0; i<size(pl); i++) pl[i]->save(b);
  }

void TPlayground::load(TBuffer& b) {
  cleanup();

  for(int o=0; o<size(objects); o++) {delete objects[o]; objects[o] = NULL;}
  
  b.get(time);
  b.get(kqpos);
  b.get(seed);
  b.get(sizeX);
  b.get(sizeY);

  for(int y=0; y<sizeY; y++) for(int x=0; x<sizeX; x++) {
    b.getC(map[y][x]);
    b.get(owner[y][x]);
    }
  
  int ii = b.getInt();
  objects.resize(ii);
  for(int o=0; o<ii; o++) objects[o] = b.getObject();

  ii = b.getInt();
  pl.resize(ii);
  for(int i=0; i<ii; i++) {
    TPlayer *pla = new TPlayer;
    pla->load(b);
    pla->index = i;
    pl[i] = pla;
    }
  }

void TPlayer::save(TBuffer& b) {
  b.put(lives);
  b.put(frags);
  b.put(score);
  b.put(level);
  b.put(ammo);
  b.put(typeval);
  b.put(typenum);
  b.put(color);
  for(int k=0; k<8; k++) b.put(key[k]);
  b.put(spx);
  b.put(spy);
  }

void TPlayer::load(TBuffer& b) {
  b.get(lives);
  b.get(frags);
  b.get(score);
  b.get(level);
  b.get(ammo);
  b.get(typeval);
  b.get(typenum);
  b.get(color);
  for(int k=0; k<8; k++) b.get(key[k]);
  b.get(spx);
  b.get(spy);
  createIcons();
  }

// message types

#define mAllocatePlayers 1 // new players are allocated (colors supplied)
#define mPlayground      2 // total playground and keypress information
#define mKeypressInfo    3 // keypress information
#define mSynchronizer    4 // synchronizer (not used)
#define mNotAccepted     5 // we don't want you here

void initBoard() {
  d.setCurrentTime(0);
  g.init(c.sizeX, c.sizeY);

  kqend = 1;
  for(int u=0; u<KQSIZE; u++) kqueue[u].time = NEVER;
  kqueue[0].time = -NEVER;
  }

void initPlayers() {
  for(int i=0; i<size(lps); i++) {
    TPlayer *p = new TPlayer;
    p->init(lps[i].color);
    lps[i].index = g.addPlayer(p);
    }
  }

IPaddress serverIP;
TCPsocket mainSocket;
vector<TCPsocket> clients;
SDLNet_SocketSet set;

void initServer() {
  if(SDLNet_Init()==-1)
    throw TException("SDLNet Init", 0, SDLNet_GetError());

  if(SDLNet_ResolveHost(&serverIP,NULL, c.port)==-1)
    throw TException("Resolve Host", 0, SDLNet_GetError());

  mainSocket=SDLNet_TCP_Open(&serverIP);

  if(!mainSocket)
    throw TException("Accept", 0, SDLNet_GetError());
  
  set=SDLNet_AllocSocketSet(64);
  }

void initClient() {
  if(SDLNet_Init()==-1)
    throw TException("SDLNet Init", 0, SDLNet_GetError());
      
  if(SDLNet_ResolveHost(&serverIP, c.server.c_str(), c.port)==-1)
    throw TException("Resolve Host", 0, SDLNet_GetError());

  for(int k=0;; k++) {
    mainSocket = SDLNet_TCP_Open(&serverIP);
    if(mainSocket) break;
    SDL_Delay(500);
    if(k == 20) throw TException("Could not connect");
    }
  int t0 = d.getCurrentTime();
  
  set=SDLNet_AllocSocketSet(1);
  if(SDLNet_TCP_AddSocket(set, mainSocket)==-1)
    throw TException("AddSocket", 0, SDLNet_GetError());

  TBuffer buf;
 
  buf.receive(mainSocket);
  int t1 = d.getCurrentTime();
  int i = buf.getInt();
  if(i == mNotAccepted)
    throw TException("Not accepted");
  if(i != mPlayground)
    throw TException("Bad message type");

  g.load(buf);

  for(int u=0; u<KQSIZE; u++) kqueue[u].time = NEVER;
  buf.get(kqend);
  kqueue[(kqend-1)%KQSIZE].time = -NEVER;
  int p = g.kqpos;
  while(p != kqend) {
    buf.get(kqueue[p%KQSIZE]);
    p++;
    }
  
  int ctime = buf.getInt();
  ctime += (t1-t0)/2;
  d.setCurrentTime(ctime);
  printf("Latency %d, time is %d (know from %d)\n", t1-t0, ctime, g.time);

  for(int p=0; p<size(lps); p++) {
    TKeypress kp;
    kp.time = ctime+1;
    kp.key = keyJoinGame;
    kp.player = p;
    addKeyToQueue(kp);
    kp.player = 0x1000000 | lps[p].color;
    broadcastKey(kp);
    lps[p].index = 0; // yet unassigned
    }
  }

void initSaves() {
  firstSave = 0;
  lastSave = 0;
  saves[0].buf.clear();
  g.save(saves[0]);
  }

void closeServer() {
  if(mainSocket) SDLNet_TCP_Close(mainSocket);
  for(int k=0; k<clients.size(); k++)
    SDLNet_TCP_Close(clients[k]);
  clients.resize(0);
  if(set) {
    SDLNet_FreeSocketSet(set);
    set = NULL;
    }
  }

void closeClient() {
  if(mainSocket) SDLNet_TCP_Close(mainSocket);
  if(set) {
    SDLNet_FreeSocketSet(set);
    set = NULL;
    }
  }

void startGame() {
  if(!c.ok) return;
  switch(c.gt) {
    case gLocal:      
      d.initForGame();
      initBoard();
      initPlayers();
      d.loop();
      break;
    
    case gHost:
      try {
        initServer();
        d.initForGame();
        initBoard();
        initPlayers();
        initSaves();
        d.loop();
        closeServer();
        }
      catch(TException e) {
        c.err = e.msg;
        if(e.ds != "") c.err += ": " + e.ds;
        printf("%s/%d/%s\n", e.msg.c_str(), e.data, e.ds.c_str());
        closeServer();
        }
      break;
    
    case gClient:
      try {
        initClient();
        d.initForGame();
        for(int i=0; i<size(g.pl); i++) if(g.pl[i]) g.pl[i]->createIcons();
        initSaves();
        d.loop();
        closeClient();
        }
      catch(TException e) {
        c.err = e.msg;
        if(e.ds != "") c.err += ": " + e.ds;
        printf("%s/%d/%s\n", e.msg.c_str(), e.data, e.ds.c_str());
        closeClient();
        }
      break;
    }
  }

void dataServer() {
  while(size(clients)) {
    int numready=SDLNet_CheckSockets(set, 0);
    if(!numready) break;
    if(numready == -1)
      throw TException("CheckSockets", size(clients), SDLNet_GetError());
    
    for(int p=0; p<size(clients); p++) {
      TCPsocket skt = clients[p];
      if(SDLNet_SocketReady(skt)) {
        TBuffer buf;
        bool b = buf.receive(skt);
        if(!b) {
          printf("Client disconnected\n");
          SDLNet_TCP_DelSocket(set, skt);
          SDLNet_TCP_Close(skt);
          clients[p] = clients[size(clients)-1];
          clients.resize(size(clients)-1);
          p--;
          continue;
          }
        switch(buf.getInt()) {
          case mKeypressInfo: {
            TKeypress kp;
            buf.get(kp);
            addKeyToQueue(kp);
            for(int q=0; q<size(clients); q++) if(q!=p) buf.send(clients[q]);
            break;
            }
  
          case mSynchronizer:
            // NIY
            break;
          
          default:
            throw TException("Unknown message");
          }
        }
      }
    }

  TCPsocket skt = SDLNet_TCP_Accept(mainSocket);
  if(!skt) return;
  
  printf("Accepted a connection.\n");
  clients.push_back(skt);
  
  if(SDLNet_TCP_AddSocket(set, skt)==-1)
    throw TException("AddSocket", 0, SDLNet_GetError());
  
  TBuffer desc;
  desc.putInt(mPlayground);
  saves[firstSave].appendTo(desc);
  desc.put(kqend);
  int p = saves[firstSave].getKQPos();
  while(p != kqend) {
    desc.put(kqueue[p%KQSIZE]);
    p++;
    }
  desc.putInt(d.getCurrentTime());
  desc.send(skt);
  }

void dataClient() {
  
  if(mainSocket) while(1) { 
    int numready = SDLNet_CheckSockets(set, 0);
    if(!numready) break;
    if(numready == -1)
      throw TException("CheckSockets", 0, SDLNet_GetError());
   
    if(SDLNet_SocketReady(mainSocket)) {
      TBuffer buf;
      bool b = buf.receive(mainSocket);      
      if(!b) {
        printf("Disconnected from the server\n");
        SDLNet_TCP_Close(mainSocket);
        mainSocket = NULL;
        return;
        }
      switch(buf.getInt()) {
        case mKeypressInfo: {
          TKeypress kp;
          buf.get(kp);
          addKeyToQueue(kp);
          break;
          }

        case mSynchronizer:
          // NIY
          break;
        
        default:
          throw TException("Unknown message");
        }
      }
    }
  }

void makeBackups() {
  if(c.gt != gLocal)
  if(g.time > saves[lastSave].getTime() + SAVEPERIOD) {
    lastSave++;
    lastSave %= QSAVES;
    if(lastSave == firstSave) {
      firstSave++;
      firstSave %= QSAVES;
      }
    saves[lastSave].buf.clear();
    g.save(saves[lastSave]);
    // printf("Backup at %d, size %d\n", lastSave, size(saves[lastSave].buf));\
    // TEST
    // saves[lastSave].readpos = 0;
    // g.load(saves[lastSave]);
    }
  }

void dataTransfer() {
  makeBackups();
  switch(c.gt) {
    case gLocal:
      break;
    
    case gHost:
      dataServer();
      break;
    
    case gClient:
      dataClient();
      break;
    }
  }

void addKeyToQueue(TKeypress& k) {
  if(k.time <= g.time) {
    int steps = 0;
    while(saves[lastSave].getTime() >= k.time) {
      if(lastSave == firstSave) 
        throw TException("Retraction impossible");
      if(lastSave == 0) lastSave = QSAVES-1;
      else lastSave--;
      steps++;
      }
    saves[lastSave].readpos = 0;
    g.load(saves[lastSave]);
    }
  int pos = kqend;
  kqend++;
  int pos2;
  while(before(k, kqueue[(pos-1)%KQSIZE])) {
    kqueue[pos%KQSIZE] = kqueue[(pos-1)%KQSIZE];
    pos--;
    }
  kqueue[pos%KQSIZE] = k;
  }

void broadcastKey(TKeypress& kp) {
  if(c.gt == gLocal) return;
  TBuffer buf;
  buf.putInt(mKeypressInfo);
  buf.put(kp);
  
  if(c.gt == gClient) {
    if(mainSocket) buf.send(mainSocket);
    }
  else {
    for(int i=0; i<size(clients); i++) buf.send(clients[i]);
    }
  }

void bufferTest() {
  TBuffer buf;
  g.save(buf);
  buf.readpos = 0;
  g.load(buf);
  }

==> untahris-3.1-src.tgz: untahris/net.cpp <==
// Untahris (Common Playground)
// copyright (C) 2006 longuens@users.sourceforge.net
// Released under GNU General Public License (see the file COPYING)
// --------------------------------------------------------------
// Network utilities.

// The algorithm for netplay is as follows. Each keypress (or join) is
// stored in the key queue, together with its time. These are sent
// across the network. Tha playground acts deterministically given
// these keypresses. Backups of playground are saved, so if a keypress
// arrives when the playground is already newer than its time, we can
// retrace to the old playground.

#include <SDL/SDL_net.h>

// struct to be sent via net, and for storing backups
struct TBuffer {
  int readpos;
  vector<char> buf;
  
  template<class T> void put(T t) {
    const int s = sizeof(T);
    char *p = (char*) &t;
    for(int i=0; i<s; i++) buf.push_back(p[i]);
    }

  void putChar(char c) { put(c); }
  void putInt(int32 c) { put(c); }
  
  int getInt() { int32 i; get(i); return i;}
  
  void putC(CellType c) { putChar((char) c); }
  void getC(CellType& c) {char ch; get(ch); c = CellType(ch); }
  
  template<class T> void get(T& t) {
    if(readpos >= size(buf)) {
      printf("Protocol error\n");
      exit(3);
      }
    const int s = sizeof(T);
    char *p = (char*) &t;
    for(int i=0; i<s; i++) p[i] = buf[readpos++];
    }
  
  TPlayObject *getObject();
  
  bool send(TCPsocket skt);
  bool receive(TCPsocket skt);
  
  int getTime() {
    return *((int*)(&(buf[0])));
    }

  int getKQPos() {
    return *((int*)(&(buf[4])));
    }
  
  void appendTo(TBuffer& b) {
    for(int i=0; i<size(buf); i++) b.putChar(buf[i]);
    }
  };

#define QSAVES 32
#define SAVEPERIOD 500

int lastSave;
int firstSave;

TBuffer saves[QSAVES];

bool TBuffer::send(TCPsocket skt) {
  int result;
  
  int len = size(buf);
  
  result = SDLNet_TCP_Send(skt, &len, 4);
  if(result != 4) return false;

  result = SDLNet_TCP_Send(skt, &(buf[0]), len);
  if(result != len) return false;
  return true;
  }

bool TBuffer::receive(TCPsocket skt) {
  int result, len;

  result = SDLNet_TCP_Recv(skt, &len, 4);
  if(result != 4) return false;
  buf.resize(len);

  int pos = 0;
  while(pos < len) {
    result = SDLNet_TCP_Recv(skt, &(buf[pos]), len-pos);
    if(result < 1) return false;
    pos += result;
    }

  readpos = 0;
  return true;
  }

void TPlayground::save(TBuffer& b) {
  b.put(time);
  b.put(kqpos);
  b.put(paused);
  b.put(seed);
  b.put(sizeX);
  b.put(sizeY);
  
  for(int y=0; y<sizeY; y++) for(int x=0; x<sizeX; x++) {
    b.putC(map[y][x]);
    b.put(owner[y][x]);
    }
  
  b.putInt(size(objects));
  for(int o=0; o<size(objects); o++) objects[o]->save(b);

  b.putInt(size(pl));
  for(int i=0; i<size(pl); i++) pl[i]->save(b);
  }

void TPlayground::load(TBuffer& b) {
  cleanup();

  for(int o=0; o<size(objects); o++) {delete objects[o]; objects[o] = NULL;}
  
  b.get(time);
  b.get(kqpos);
  b.get(paused);
  b.get(seed);
  b.get(sizeX);
  b.get(sizeY);

  for(int y=0; y<sizeY; y++) for(int x=0; x<sizeX; x++) {
    b.getC(map[y][x]);
    b.get(owner[y][x]);
    }
  
  int ii = b.getInt();
  objects.resize(ii);
  for(int o=0; o<ii; o++) objects[o] = b.getObject();

  ii = b.getInt();
  pl.resize(ii);
  for(int i=0; i<ii; i++) {
    TPlayer *pla = new TPlayer;
    pla->load(b);
    pla->index = i;
    pl[i] = pla;
    }
  }

void TPlayer::save(TBuffer& b) {
  b.put(lives);
  b.put(frags);
  b.put(score);
  b.put(level);
  b.put(ammo);
  b.put(typeval);
  b.put(editing); b.put(charid); b.put(editid);
  b.put(color);
  for(int k=0; k<8; k++) b.put(key[k]);
  b.put(killer);
  }

void TPlayer::load(TBuffer& b) {
  b.get(lives);
  b.get(frags);
  b.get(score);
  b.get(level);
  b.get(ammo);
  b.get(typeval);
  b.get(editing); b.get(charid); b.get(editid);
  b.get(color);
  for(int k=0; k<8; k++) b.get(key[k]);
  b.get(killer);
  createIcons();
  }

// message types

#define mAllocatePlayers 1 // new players are allocated (colors supplied)
#define mPlayground      2 // total playground and keypress information
#define mKeypressInfo    3 // keypress information
#define mSynchronizer    4 // synchronizer (not used)
#define mNotAccepted     5 // we don't want you here

void initBoard() {
  d.setCurrentTime(0);
  
  g.init(
    (d.sizeX * c.pctX) / 100 / c.res,
    ((d.sizeY - 20 * c.scorelines) * c.pctY) / 100 / c.res
    );

  kqend = 1;
  for(int u=0; u<KQSIZE; u++) kqueue[u].time = NEVER;
  kqueue[0].time = -NEVER;
  }

void initPlayers() {
  for(int i=0; i<size(lps); i++) {
    TPlayer *p = new TPlayer;
    p->init(lps[i].color);
    lps[i].index = g.addPlayer(p);
    }
  }

IPaddress serverIP;
TCPsocket mainSocket;
vector<TCPsocket> clients;
SDLNet_SocketSet set;

void initServer() {
  if(SDLNet_Init()==-1)
    throw TException("SDLNet Init", 0, SDLNet_GetError());

  if(SDLNet_ResolveHost(&serverIP,NULL, c.port)==-1)
    throw TException("Resolve Host", 0, SDLNet_GetError());

  mainSocket=SDLNet_TCP_Open(&serverIP);

  if(!mainSocket)
    throw TException("Accept", 0, SDLNet_GetError());
  
  set=SDLNet_AllocSocketSet(64);
  }

void initClient() {
  if(SDLNet_Init()==-1)
    throw TException("SDLNet Init", 0, SDLNet_GetError());
      
  if(SDLNet_ResolveHost(&serverIP, c.server.c_str(), c.port)==-1)
    throw TException("Resolve Host", 0, SDLNet_GetError());

  for(int k=0;; k++) {
    mainSocket = SDLNet_TCP_Open(&serverIP);
    if(mainSocket) break;
    SDL_Delay(500);
    if(k == 20) throw TException("Could not connect");
    }
  int t0 = d.getCurrentTime();
  
  set=SDLNet_AllocSocketSet(1);
  if(SDLNet_TCP_AddSocket(set, mainSocket)==-1)
    throw TException("AddSocket", 0, SDLNet_GetError());

  TBuffer buf;
 
  buf.receive(mainSocket);
  int t1 = d.getCurrentTime();
  int i = buf.getInt();
  if(i == mNotAccepted)
    throw TException("Not accepted");
  if(i != mPlayground)
    throw TException("Bad message type");
    
  buf.get(gc);

  g.load(buf);

  for(int u=0; u<KQSIZE; u++) kqueue[u].time = NEVER;
  buf.get(kqend);
  kqueue[(kqend-1)%KQSIZE].time = -NEVER;
  int p = g.kqpos;
  while(p != kqend) {
    buf.get(kqueue[p % KQSIZE]);
    p++;
    }
  
  int ctime = buf.getInt();
  ctime += (t1-t0)/2;
  d.setCurrentTime(ctime);
  printf("Latency %d, time is %d (know from %d)\n", t1-t0, ctime, g.time);

  for(int p=0; p<size(lps); p++) {
    TKeypress kp;
    kp.time = ctime+1;
    kp.key = keyJoinGame;
    kp.player = p;
    addKeyToQueue(kp);
    kp.player = 0x1000000 | lps[p].color | (0x2000000 * p);
    broadcastKey(kp);
    lps[p].index = 0; // yet unassigned
    }
  }

void initSaves() {
  firstSave = 0;
  lastSave = 0;
  saves[0].buf.clear();
  g.save(saves[0]);
  }

void closeServer() {
  if(mainSocket) SDLNet_TCP_Close(mainSocket);
  for(int k=0; k<clients.size(); k++)
    SDLNet_TCP_Close(clients[k]);
  clients.resize(0);
  if(set) {
    SDLNet_FreeSocketSet(set);
    set = NULL;
    }
  }

void closeClient() {
  if(mainSocket) SDLNet_TCP_Close(mainSocket);
  if(set) {
    SDLNet_FreeSocketSet(set);
    set = NULL;
    }
  }

void startGame() {
  if(!c.ok) return;
  switch(c.gt) {
    case gLocal:      
      d.initForGame();
      initBoard();
      d.initBackground();
      initPlayers();
      d.loop();
      break;
    
    case gHost:
      try {
        initServer();
        d.initForGame();
        initBoard();
        d.initBackground();
        initPlayers();
        initSaves();
        d.loop();
        closeServer();
        }
      catch(TException e) {
        c.err = e.msg;
        if(e.ds != "") c.err += ": " + e.ds;
        printf("%s/%d/%s\n", e.msg.c_str(), e.data, e.ds.c_str());
        closeServer();
        }
      break;
    
    case gClient:
      TGameConfig sgc = gc;
      try {
        initClient();
        d.initForGame();
        d.initBackground();
        for(int i=0; i<size(g.pl); i++) if(g.pl[i]) g.pl[i]->createIcons();
        initSaves();
        d.loop();
        closeClient();
        }
      catch(TException e) {
        c.err = e.msg;
        if(e.ds != "") c.err += ": " + e.ds;
        printf("%s/%d/%s\n", e.msg.c_str(), e.data, e.ds.c_str());
        closeClient();
        }
      gc = sgc;
      break;
    }
  }

void dataServer() {
  while(size(clients)) {
    int numready=SDLNet_CheckSockets(set, 0);
    if(!numready) break;
    if(numready == -1)
      throw TException("CheckSockets", size(clients), SDLNet_GetError());
    
    for(int p=0; p<size(clients); p++) {
      TCPsocket skt = clients[p];
      if(SDLNet_SocketReady(skt)) {
        TBuffer buf;
        bool b = buf.receive(skt);
        if(!b) {
          printf("Client disconnected\n");
          SDLNet_TCP_DelSocket(set, skt);
          SDLNet_TCP_Close(skt);
          clients[p] = clients[size(clients)-1];
          clients.resize(size(clients)-1);
          p--;
          continue;
          }
        switch(buf.getInt()) {
          case mKeypressInfo: {
            TKeypress kp;
            buf.get(kp);
            addKeyToQueue(kp);
            for(int q=0; q<size(clients); q++) if(q!=p) buf.send(clients[q]);
            break;
            }
  
          case mSynchronizer:
            // NIY
            break;
          
          default:
            throw TException("Unknown message");
          }
        }
      }
    }

  TCPsocket skt = SDLNet_TCP_Accept(mainSocket);
  if(!skt) return;
  
  printf("Accepted a connection.\n");
  clients.push_back(skt);
  
  if(SDLNet_TCP_AddSocket(set, skt)==-1)
    throw TException("AddSocket", 0, SDLNet_GetError());
  
  TBuffer desc;
  desc.putInt(mPlayground);
  desc.put(gc);
  saves[firstSave].appendTo(desc);
  desc.put(kqend);
  int p = saves[firstSave].getKQPos();
  while(p != kqend) {
    desc.put(kqueue[p % KQSIZE]);
    p++;
    }
  desc.putInt(d.getCurrentTime());
  desc.send(skt);
  }

void dataClient() {
  
  if(mainSocket) while(1) { 
    int numready = SDLNet_CheckSockets(set, 0);
    if(!numready) break;
    if(numready == -1)
      throw TException("CheckSockets", 0, SDLNet_GetError());
   
    if(SDLNet_SocketReady(mainSocket)) {
      TBuffer buf;
      bool b = buf.receive(mainSocket);      
      if(!b) {
        printf("Disconnected from the server\n");
        SDLNet_TCP_Close(mainSocket);
        mainSocket = NULL;
        return;
        }
      switch(buf.getInt()) {
        case mKeypressInfo: {
          TKeypress kp;
          buf.get(kp);
          addKeyToQueue(kp);
          break;
          }

        case mSynchronizer:
          // NIY
          break;
        
        default:
          throw TException("Unknown message");
        }
      }
    }
  }

void makeBackups() {
  if(c.gt != gLocal)
  if(g.time > saves[lastSave].getTime() + SAVEPERIOD) {
    lastSave++;
    lastSave %= QSAVES;
    if(lastSave == firstSave) {
      firstSave++;
      firstSave %= QSAVES;
      }
    saves[lastSave].buf.clear();
    g.save(saves[lastSave]);
    // printf("Backup at %d, size %d\n", lastSave, size(saves[lastSave].buf));\
    // TEST
    // saves[lastSave].readpos = 0;
    // g.load(saves[lastSave]);
    }
  }

void dataTransfer() {
  makeBackups();
  switch(c.gt) {
    case gLocal:
      break;
    
    case gHost:
      dataServer();
      break;
    
    case gClient:
      dataClient();
      break;
    }
  }

void addKeyToQueue(TKeypress& k) {
  if(k.time <= g.time) {
    int steps = 0;
    while(saves[lastSave].getTime() >= k.time) {
      if(lastSave == firstSave) 
        throw TException("Retraction impossible");
      if(lastSave == 0) lastSave = QSAVES-1;
      else lastSave--;
      steps++;
      }
    saves[lastSave].readpos = 0;
    g.load(saves[lastSave]);
    }
  int pos = kqend;
  kqend++;
  int pos2;
  while(before(k, kqueue[(pos-1)%KQSIZE])) {
    kqueue[pos%KQSIZE] = kqueue[(pos-1)%KQSIZE];
    pos--;
    }
  kqueue[pos%KQSIZE] = k;
  }

void broadcastKey(TKeypress& kp) {
  if(c.gt == gLocal) return;
  TBuffer buf;
  buf.putInt(mKeypressInfo);
  buf.put(kp);
  
  if(c.gt == gClient) {
    if(mainSocket) buf.send(mainSocket);
    }
  else {
    for(int i=0; i<size(clients); i++) buf.send(clients[i]);
    }
  }
